from typing import Dict, Union, List, Optional
import os
from xml.etree import ElementTree

from sfml import sf

from ..reslib.resource_path import split_path


class Tileset:
    """ Base class for tilesets, called when loading a tileset """
    def __init__(self, xml: ElementTree.Element, path: str):
        self.xml_root: ElementTree.Element = xml
        self.path = path

        image = self.xml_root.find('image')

        self.name: str = self.xml_root.get('name')
        self.columns: int = int(self.xml_root.get('columns'))
        self.rows: int = int(image.get('height')) // int(self.xml_root.get('tileheight'))
        self.tile_width: int = int(self.xml_root.get('tilewidth'))
        self.tile_height: int = int(self.xml_root.get('tileheight'))
        self.tile_count: int = int(self.xml_root.get('tilecount'))
        self.texture_source = image.get('source')
        self.texture: Optional[sf.Texture] = None

        self.properties: Dict[int, Dict[str, Union[bool, int, float, str, sf.Color]]] = {}
        self.animations: Dict[int, List[Dict[str, int]]] = {}

    def load(self):
        texture_path = split_path(os.path.dirname(self.path)) + split_path(os.path.splitext(self.texture_source)[0])

        from ..reslib import Res
        res = Res
        if texture_path[0] == "assets":
            texture_path.pop(0)
        for folder in texture_path:
            res = getattr(res, folder)

        self.texture = res

        for tile in self.xml_root.findall('tile'):
            tile_id = int(tile.get('id'))
            for child in tile:
                if child.tag == "properties":
                    if tile_id not in self.properties:
                        self.properties[tile_id] = {}

                    for prop in child:
                        name = prop.get('name')
                        prop_type = prop.get('type')
                        val = prop.get('value')
                        if prop_type == "bool":
                            val = True if val == "true" else False
                        elif prop_type == "int":
                            val = int(val)
                        elif prop_type == "float":
                            val = float(val)
                        elif prop_type == "color":
                            val = sf.Color(int(val[1:3], 16), int(val[3:5], 16), int(val[5:7], 16), int(val[7:9], 16))
                        self.properties[tile_id].update({name: val})

                elif child.tag == "animation":
                    if tile_id not in self.animations:
                        self.animations[tile_id] = []

                    for frame in child:
                        self.animations[tile_id].append(
                            {
                                'id': int(frame.get('tileid')),
                                'duration': int(frame.get('duration'))
                            }
                        )

    def get_tile_tex_coord(self, id: int):
        if id < self.tile_count:
            tx = (id % self.columns) * self.tile_width
            ty = (id // self.columns) * self.tile_height
            return sf.Vector2(tx, ty)
        else:
            raise IndexError(f'Tile id {id} not found in tileset {self.name}.')


class MapTileset:
    """ When loading a TiledMap, a MapTileset is created for each tileset used in the map """
    def __init__(self, tileset: Tileset, firstgid: int):
        self._data: Tileset = tileset
        self._first_gid: int = firstgid
        self._last_gid = self.first_gid + self._data.tile_count - 1

    @property
    def name(self) -> str:
        return self._data.name

    @property
    def columns(self) -> int:
        return self._data.columns

    @property
    def rows(self) -> int:
        return self._data.rows

    @property
    def tile_width(self) -> int:
        return self._data.tile_width

    @property
    def tile_height(self) -> int:
        return self._data.tile_height

    @property
    def tile_count(self) -> int:
        return self._data.tile_count

    @property
    def first_gid(self) -> int:
        return self._first_gid

    @property
    def last_gid(self) -> int:
        return self._last_gid

    @property
    def texture(self) -> sf.Texture:
        return self._data.texture

    @property
    def properties(self) -> Dict[int, Dict[str, Union[bool, int, float, str, sf.Color]]]:
        return self._data.properties

    @property
    def animations(self) -> Dict[int, List[Dict[str, int]]]:
        return self._data.animations

    def get_tile_tex_coord(self, id: int):
        return self._data.get_tile_tex_coord(id)



from sfml import sf

from ..data.callbacks import callback, HasCallbacks
from ..data.game_obj import GameObject
from .camera import Camera


class RenderWindow(GameObject, HasCallbacks, sf.RenderWindow):
    def __init__(self, videomode: sf.VideoMode, title: str, style: sf.Style = sf.Style.DEFAULT):
        super().__init__(videomode, title, style)
        self.style = style
        self.base_videomode = videomode
        self.base_size = sf.Vector2(videomode.width, videomode.height)
        self.base_title = title
        self._ui_view = Camera("ui_view", -1)
        self._ui_view.reset((0, 0), (self.game.V_WIDTH, self.game.V_HEIGHT))
        self._ui_view.reset_viewport((0, 0), (1, 1))

    @property
    def ui_view(self):
        return self._ui_view

    @callback("on_close")
    def on_close(self, user_fn):
        return user_fn

    def close(self):
        super().close()
        self.callbacks.call("on_close")

    def on_resize(self):
        self.game.scale_view()



from sfml import sf

from ..data.game_obj import GameObject
from . import transitions


class Layer(GameObject, sf.Drawable):
    """
    A Layer is a collection of Drawables. It can be drawn on the window.
    Used to organize the order of drawing in the game.
    """
    def __init__(self, name: str, *args: sf.Drawable):
        super().__init__()
        self.name = name
        self._drawables = []
        for arg in args:
            if isinstance(arg, sf.Drawable):
                self._drawables.append(arg)
            else:
                raise TypeError("Layer can only contain Drawables.")

    def add(self, *args: sf.Drawable):
        for arg in args:
            if isinstance(arg, sf.Drawable):
                self._drawables.append(arg)
            else:
                raise TypeError("You can only add Drawables to Layer")

    def remove(self, drawable: sf.Drawable):
        if drawable in self._drawables:
            self._drawables.remove(drawable)

    def ysort(self):
        """Sort all drawables of the layer by y position
        """
        self._drawables.sort(key=lambda x: x.position.y)

    def update(self):
        to_remove = []
        for spr in self._drawables:
            if isinstance(spr, transitions.Transition):
                if spr.ended:
                    to_remove.append(spr)
        for tr in to_remove:
            self._drawables.remove(tr)

    def draw(self, target, states):
        for drawable in self._drawables:
            if isinstance(drawable, transitions.Transition):
                if drawable.started:
                    target.draw(drawable, states)
            else:
                target.draw(drawable, states)

    def __iter__(self):
        return iter(self._drawables)


class Mask(GameObject, sf.Drawable):
    def __init__(self, name: str, width: int, height: int, fill_color: sf.Color, *args: sf.Drawable):
        super().__init__()
        self.name = name
        self.fill_color = fill_color
        self.render_texture = sf.RenderTexture(width, height)
        self.sprite = sf.Sprite(self.render_texture.texture)
        self._drawables = []
        for arg in args:
            if isinstance(arg, sf.Drawable):
                self._drawables.append(arg)
            else:
                raise TypeError("Mask can only contain Drawables.")

    def add(self, *args: sf.Drawable):
        for arg in args:
            if isinstance(arg, sf.Drawable):
                self._drawables.append(arg)
            else:
                raise TypeError("Mask can only contain Drawables.")

    def update(self):
        self.render_texture.clear(self.fill_color)
        temp_texture = sf.RenderTexture(self.render_texture.size.x, self.render_texture.size.y)
        temp_texture.clear(self.fill_color)
        for drawable in self._drawables:
            temp_texture.draw(drawable, sf.RenderStates(sf.BLEND_MULTIPLY))
        temp_texture.display()
        self.render_texture.draw(sf.Sprite(temp_texture.texture), sf.RenderStates(sf.BLEND_MULTIPLY))
        self.render_texture.display()
        self.sprite = sf.Sprite(self.render_texture.texture)

    def draw(self, target, states):
        target.draw(self.sprite, states)

    def __iter__(self):
        return iter(self._drawables)



import math
import random

from sfml import sf

from ..data.callbacks import HasCallbacks as __HasCallbacks
from ..data.callbacks import callback as _callback
from ..data.game_obj import GameObject as __GameObject


class Transition(__GameObject, __HasCallbacks, sf.Drawable):
    def __init__(self):
        if self.__class__.__name__ == __class__.__name__:
            raise NotImplementedError("Transition class is not instantiable. Please use inheritance to create a Transition.")
        super().__init__()
        self.render_texture = sf.RenderTexture(self.game.window.ui_view.size.x, self.game.window.ui_view.size.y)
        self.r = self.g = self.b = 0
        self.a = 255
        self.blend_mode = sf.BLEND_NONE
        self.shapes = []
        self.sprite = sf.Sprite(self.render_texture.texture)
        self.ended = False
        self.started = False

    def reset(self):
        self.render_texture = sf.RenderTexture(self.game.window.ui_view.size.x, self.game.window.ui_view.size.y)
        self.r = self.g = self.b = 0
        self.a = 255
        self.blend_mode = sf.BLEND_NONE
        self.shapes = []
        self.sprite = sf.Sprite(self.render_texture.texture)
        self.ended = False
        self.started = False

    @property
    def width(self):
        return self.render_texture.size.x

    @property
    def height(self):
        return self.render_texture.size.y

    @property
    def fill_color(self):
        return sf.Color(self.r, self.g, self.b, self.a)

    @fill_color.setter
    def fill_color(self, value: sf.Color):
        self.r = value.r
        self.g = value.g
        self.b = value.b
        self.a = value.a

    @_callback("on_start")
    def on_start(self, user_fn):
        return user_fn

    def start(self):
        if not self.started:
            self.started = True
            self.game.transitions.append(self)
            self.callbacks.call("on_start")

    @_callback("on_end")
    def on_end(self, user_fn):
        return user_fn

    def end(self):
        if not self.ended:
            self.ended = True
            self.game.transitions.remove(self)
            self.callbacks.call("on_end")

    @staticmethod
    def updater(update_func):
        def _decorate(self):
            if not self.ended:
                self.render_texture.clear(self.fill_color)
                if self.started:
                    update_func(self)
                for shape in self.shapes:
                    self.render_texture.draw(shape, sf.RenderStates(self.blend_mode))
                self.render_texture.display()
                self.sprite.texture = self.render_texture.texture
        return _decorate

    def update(self):
        pass

    def draw(self, target, transformations):
        target.draw(self.sprite)


class FadeIn(Transition):
    """ Fade transition from black screen to transparent"""
    def __init__(self, speed=5):
        super().__init__()
        self.speed = speed
        self.reset()

    def reset(self):
        super().reset()
        self.a = 255
        self.update()

    @Transition.updater
    def update(self):
        self.a -= self.speed
        if self.a <= 0:
            self.a = 0
            self.end()


class FadeOut(Transition):
    """ Fade transition from transparent to black screen"""
    def __init__(self, speed=5):
        super().__init__()
        self.speed = speed
        self.reset()

    def reset(self):
        super().reset()
        self.a = 0
        self.update()

    @Transition.updater
    def update(self):
        self.a += self.speed
        if self.a >= 255:
            self.a = 255
            self.end()


class CircleOpen(Transition):
    def __init__(self, speed=5):
        super().__init__()
        self.speed = speed
        self.limit = math.sqrt((self.width/2)**2 + (self.height/2)**2)
        self.reset()

    def reset(self):
        super().reset()
        shape = sf.CircleShape(0.1)
        shape.origin = (0.1, 0.1)
        shape.position = (self.width / 2, self.height / 2)
        shape.fill_color = sf.Color.TRANSPARENT
        self.shapes.append(shape)
        self.update()

    @Transition.updater
    def update(self):
        if self.shapes[0].radius > self.limit:
            self.end()
        elif self.shapes[0].radius <= self.limit:
            self.shapes[0].radius += self.speed
        else:
            self.shapes[0].radius = 0
        self.shapes[0].origin = (self.shapes[0].radius, self.shapes[0].radius)


class CircleClose(Transition):
    """ Closing circle transition"""
    def __init__(self, speed=5):
        super().__init__()
        self.speed = speed
        self.reset()

    def reset(self):
        super().reset()
        shape = sf.CircleShape(math.sqrt((self.width/2)**2 + (self.height/2)**2))
        shape.origin = (shape.radius, shape.radius)
        shape.position = (self.width / 2, self.height / 2)
        shape.fill_color = sf.Color.TRANSPARENT
        self.shapes.append(shape)
        self.update()

    @Transition.updater
    def update(self):
        if self.shapes[0].radius == 0:
            self.end()
        elif self.shapes[0].radius >= self.speed:
            self.shapes[0].radius -= self.speed
        else:
            self.shapes[0].radius = 0
        self.shapes[0].origin = (self.shapes[0].radius, self.shapes[0].radius)


class RotatingSquareOpen(Transition):
    def __init__(self, speed=5):
        super().__init__()
        self.speed = speed
        self.limit = math.sqrt((self.width/2)**2 + (self.height/2)**2)
        self.shape = sf.RectangleShape()
        self.shape.position = (self.width/2, self.height/2)
        self.shape.size = (0.1, 0.1)
        self.shape.origin = (0.05, 0.05)
        self.shape.fill_color = sf.Color.TRANSPARENT
        self.shapes.append(self.shape)
        self.update()

    @Transition.updater
    def update(self):
        if self.shape.size.x/2 > self.limit:
            self.end()
        elif self.shape.size.x/2 <= self.limit:
            self.shape.size += sf.Vector2(self.speed, self.speed)
            self.shape.rotate(self.speed*2)
        self.shape.origin = (self.shape.size.x/2, self.shape.size.y/2)


class RotatingSquareClose(Transition):
    def __init__(self, speed: int = 5):
        super().__init__()
        self.speed = speed
        self.reset()

    def reset(self):
        super().reset()
        shape = sf.RectangleShape()
        shape.position = (self.width / 2, self.height / 2)
        shape.size = (self.width, self.width)
        shape.origin = (self.width / 2, self.width / 2)
        shape.fill_color = sf.Color.TRANSPARENT
        self.shapes.append(shape)
        self.update()

    @Transition.updater
    def update(self):
        if self.shapes[0].size.x == 0:
            self.end()
        elif self.shapes[0].size.x > self.speed:
            self.shapes[0].size -= sf.Vector2(self.speed, self.speed)
            self.shapes[0].rotate(self.speed*2)
        else:
            self.shapes[0].size = (0, 0)
        self.shapes[0].origin = (self.shapes[0].size.x/2, self.shapes[0].size.y/2)


class PixelsIn(Transition):
    """ From black screen to transparent square by square """
    def __init__(self, speed: int, pixelsize: float):
        """ Warning : a small pixelsize and big speed will lead to severe frame drops.
            Avoid using a pixelsize smaller than 8.
        """
        super().__init__()
        self.speed = speed
        self.pixel_size = pixelsize
        self.remaining = []
        self.reset()

    def reset(self):
        super().reset()
        self.remaining = []
        col_nb = math.ceil(self.render_texture.size.x / self.pixel_size)
        row_nb = math.ceil(self.render_texture.size.y / self.pixel_size)
        for x in range(col_nb):
            for y in range(row_nb):
                s = sf.RectangleShape()
                s.position = (x * self.pixel_size, y * self.pixel_size)
                s.size = (self.pixel_size, self.pixel_size)
                s.fill_color = sf.Color.BLACK
                self.shapes.append(s)
                self.remaining.append(s)
        self.update()

    @Transition.updater
    def update(self):
        if self.remaining:
            for i in range(self.speed):
                s = random.choice(self.remaining)
                s.fill_color = sf.Color.TRANSPARENT
                self.remaining.remove(s)
                if not self.remaining:
                    break
        else:
            self.end()


class PixelsOut(Transition):
    """ From transparent screen to black screen square by square """
    def __init__(self, speed: int, pixelsize: float):
        """ Warning : a small pixelsize and big speed will lead to severe frame drops.
            Avoid using a pixelsize smaller than 8.
        """
        super().__init__()
        self.speed = speed
        self.pixel_size = pixelsize
        self.shapes = []
        self.remaining = []

    def reset(self):
        super().reset()
        self.remaining = []
        col_nb = math.ceil(self.render_texture.size.x / self.pixel_size)
        row_nb = math.ceil(self.render_texture.size.y / self.pixel_size)
        for x in range(col_nb):
            for y in range(row_nb):
                s = sf.RectangleShape()
                s.position = (x * self.pixel_size, y * self.pixel_size)
                s.size = (self.pixel_size, self.pixel_size)
                s.fill_color = sf.Color.TRANSPARENT
                self.shapes.append(s)
                self.remaining.append(s)
        self.update()

    @Transition.updater
    def update(self):
        if self.remaining:
            for i in range(self.speed):
                s = random.choice(self.remaining)
                s.fill_color = sf.Color.BLACK
                self.remaining.remove(s)
                if not self.remaining:
                    break
        else:
            self.end()



import sys
import os

ABSOLUTE_PATH = sys.path[0]
RES_DIR = "assets"


def find_resource(relative_path: str):
    try:  # for release : pyInstaller stores assets in a temporary folder _MEIPASS
        absolute_path = sys._MEIPASS
    except:  # for developement
        absolute_path = ""
    return os.path.join(absolute_path, relative_path)


def split_path(path: str) -> list:
    sep = os.sep if os.sep in path else '/'
    return path.split(sep)


def normalize_path(path: str) -> str:
    path = split_path(path)
    while '..' in path:
        index = path.index('..')
        if index != 0:
            path = [p for i, p in enumerate(path) if i != index and i != index - 1]
    return os.sep.join(path)



from typing import List, Optional

from sfml import sf

from .tilesets import MapTileset


class TileTransformation:
    HORIZONTAL_FLIP = 0x80000000
    VERTICAL_FLIP = 0x40000000
    DIAGONAL_FLIP = 0x20000000
    ROT90 = HORIZONTAL_FLIP + DIAGONAL_FLIP
    ROT180 = HORIZONTAL_FLIP + VERTICAL_FLIP
    ROT270 = VERTICAL_FLIP + DIAGONAL_FLIP

    @staticmethod
    def get_transformations(tile_id: int):
        transformations = []
        if bool((tile_id & TileTransformation.HORIZONTAL_FLIP) >> 31):
            transformations.append(TileTransformation.HORIZONTAL_FLIP)
        if bool((tile_id & TileTransformation.VERTICAL_FLIP) >> 30):
            transformations.append(TileTransformation.VERTICAL_FLIP)
        if bool((tile_id & TileTransformation.DIAGONAL_FLIP) >> 29):
            transformations.append(TileTransformation.DIAGONAL_FLIP)
        return transformations


class Tile(sf.Drawable):
    def __init__(self, tilelayer, x: int, y: int, gid: int):
        super().__init__()
        self._map = tilelayer.map
        self._tileset: Optional[MapTileset] = None

        self._gid = gid
        self._id = gid

        self._transformations = TileTransformation.get_transformations(self._gid)
        for tr in self._transformations:
            self._gid -= tr
        for tileset in self._map.tilesets:
            if tileset.first_gid <= self._gid <= tileset.last_gid:
                self._id = self._gid - tileset.first_gid
                self._tileset = tileset
                break

        self._x = x*self._tileset.tile_width
        self._y = y*self._tileset.tile_height
        self._tx = (self._id % self._tileset.columns) * self._tileset.tile_width
        self._ty = (self._id // self._tileset.columns) * self._tileset.tile_height

        self._sprite: sf.VertexArray = sf.VertexArray(sf.PrimitiveType.QUADS)
        self._sprite.resize(4)

        self._sprite[0].position = self.position
        self._sprite[1].position = self.position + sf.Vector2(self._tileset.tile_width, 0)
        self._sprite[2].position = self.position + sf.Vector2(self._tileset.tile_width, self._tileset.tile_height)
        self._sprite[3].position = self.position + sf.Vector2(0, self._tileset.tile_height)

        tex_coords = self._calculate_tex_coo(self._tx, self._ty)
        self._sprite[0].tex_coords = tex_coords[0]
        self._sprite[1].tex_coords = tex_coords[1]
        self._sprite[2].tex_coords = tex_coords[2]
        self._sprite[3].tex_coords = tex_coords[3]

        self._frames = []
        self._anim_data = []
        self._anim_index = 0

        if self.id in self._tileset.animations:
            self._anim_data = self.tileset.animations[self.id]
            for frame in self._anim_data:
                vert = sf.VertexArray(sf.PrimitiveType.QUADS)
                vert.resize(4)
                tx = (frame['id'] % self._tileset.columns) * self._tileset.tile_width
                ty = (frame['id'] // self._tileset.columns) * self._tileset.tile_height
                vert[0].position = self.position
                vert[1].position = self.position + sf.Vector2(self._tileset.tile_width, 0)
                vert[2].position = self.position + sf.Vector2(self._tileset.tile_width, self._tileset.tile_height)
                vert[3].position = self.position + sf.Vector2(0, self._tileset.tile_height)
                tex_coords = self._calculate_tex_coo(tx, ty)
                vert[0].tex_coords = tex_coords[0]
                vert[1].tex_coords = tex_coords[1]
                vert[2].tex_coords = tex_coords[2]
                vert[3].tex_coords = tex_coords[3]
                self._frames.append(vert)
        self._clock = sf.Clock()

    def _calculate_tex_coo(self, tx, ty) -> List[sf.Vector2]:
        tex_coords = [
            sf.Vector2(tx, ty),
            sf.Vector2(tx + self._tileset.tile_width, ty),
            sf.Vector2(tx + self._tileset.tile_width, ty + self._tileset.tile_height),
            sf.Vector2(tx, ty + self._tileset.tile_height)
        ]
        # applying tile transformation to the texture coordinates
        modifier = [sf.Vector2(0, 0), sf.Vector2(0, 0), sf.Vector2(0, 0), sf.Vector2(0, 0)]
        index0, index1, index2, index3 = range(4)
        if TileTransformation.DIAGONAL_FLIP in self._transformations:
            index2 = 0
            index0 = 2
            tex_coords[0], tex_coords[2] = tex_coords[2], tex_coords[0]
        if TileTransformation.HORIZONTAL_FLIP in self._transformations:
            modifier[index0] += sf.Vector2(self._tileset.tile_width, 0)
            modifier[index1] += sf.Vector2(-self._tileset.tile_width, 0)
            modifier[index2] += sf.Vector2(-self._tileset.tile_width, 0)
            modifier[index3] += sf.Vector2(self._tileset.tile_width, 0)
        if TileTransformation.VERTICAL_FLIP in self._transformations:
            modifier[index0] += sf.Vector2(0, self._tileset.tile_height)
            modifier[index1] += sf.Vector2(0, self._tileset.tile_height)
            modifier[index2] += sf.Vector2(0, -self._tileset.tile_height)
            modifier[index3] += sf.Vector2(0, -self._tileset.tile_height)
        return [tex_coords[i] + modifier[i] for i in range(4)]

    def update(self):
        if self._frames:
            if self._clock.elapsed_time.milliseconds > self._anim_data[self._anim_index]['duration']:
                self._anim_index += 1
                if self._anim_index > len(self._frames) - 1:
                    self._anim_index = 0
                self._sprite = self._frames[self._anim_index]
                self._clock.restart()

    def draw(self, target, state):
        state.texture = self._tileset.texture
        target.draw(self._sprite, state)

    @property
    def gid(self) -> int:
        return self._gid

    @property
    def id(self) -> int:
        return self._id

    @property
    def animated(self):
        return len(self._frames) > 0

    @property
    def transformations(self) -> List[int]:
        return self._transformations

    @property
    def sprite(self) -> sf.VertexArray:
        return self._sprite

    @property
    def size(self) -> sf.Vector2:
        return sf.Vector2(self.tileset.tile_width, self.tileset.tile_height)

    @property
    def position(self) -> sf.Vector2:
        return sf.Vector2(self._x, self._y)

    @property
    def x(self) -> int:
        return self._x

    @property
    def y(self) -> int:
        return self._y

    @property
    def tileset(self) -> MapTileset:
        return self._tileset

    @property
    def texture(self):
        return self._tileset.texture



from typing import List, Dict
import os
from xml.etree import ElementTree

from sfml import sf

from ..data.game_obj import GameObject
from ..data.rect import Rect
from ..reslib.resource_path import find_resource
from ..reslib.tileset_manager import TilesetManager
from .tilesets import Tileset, MapTileset
from .layers import TileLayer, ObjectGroup


class TiledMap(GameObject):
    def __init__(self, path: str):
        self.name = os.path.basename(path)
        self.path = path
        self.root = ElementTree.parse(find_resource(os.path.join(os.path.dirname(self.path), self.name))).getroot()

        self._size: sf.Vector2 = sf.Vector2(int(self.root.get('width')), int(self.root.get('height')))
        self._tile_size: sf.Vector2 = sf.Vector2(int(self.root.get('tilewidth')), int(self.root.get('tileheight')))

        self.tilesets: List[MapTileset] = []
        self.layers: Dict[str, TileLayer] = {}
        self.objectgroups: Dict[str, ObjectGroup] = {}
        self._collisions = []

    def load(self):
        # loading tilesets
        for tileset_elmnt in self.root.findall('tileset'):
            name = tileset_elmnt.get('name')
            if not name:
                t = TilesetManager.get(os.path.splitext(os.path.basename(tileset_elmnt.get('source')))[0])
            else:
                t = Tileset(tileset_elmnt, self.path)
                t.load()
            self.tilesets.append(MapTileset(t, int(tileset_elmnt.get('firstgid'))))

        # loading layers
        for layer_elmnt in self.root.findall('layer'):
            layer = TileLayer(self, layer_elmnt)
            layer.parse()
            self.layers[layer.name] = layer

        # loading objectgroups
        for objectgroup_elmnt in self.root.findall('objectgroup'):
            objectgroup = ObjectGroup(self, objectgroup_elmnt)
            objectgroup.parse()
            self.objectgroups[objectgroup.name] = objectgroup

    def update(self):
        yrange = []
        xrange = []
        for cam in self.game.cameras:
            yboundsmin = max(0, int(cam.top) // 16 - 2)
            yboundsmax = min(self.size.x, int(cam.bottom) // 16 + 2)
            yrange += list(range(yboundsmin, yboundsmax))
            xboundsmin = max(0, int(cam.left) // 16 - 2)
            xboundsmax = min(self.size.y, int(cam.right) // 16 + 2)
            xrange += list(range(xboundsmin, xboundsmax))
        yrange = list(set(yrange))
        xrange = list(set(xrange))
        # iterating over the visible tiles only
        for j in yrange:
            for i in xrange:
                for layer in self.layers.values():
                    if layer.visible:
                        layer.update_tile(i, j)
        for layer in self.layers.values():
            if layer.visible:
                layer.render_texture.display()
                layer.sprite.texture = layer.render_texture.texture

    def set_collisions_objectgroup(self, name):
        if name in self.objectgroups:
            self._collisions = self.objectgroups[name].rectangles
        else:
            raise AttributeError(f"{name} is not an objectgroup of {self.name} TiledMap")

    @property
    def collisions(self) -> List[Rect]:
        return self._collisions

    @property
    def size(self) -> sf.Vector2:
        return self._size

    @property
    def width(self) -> int:
        return self._size.x

    @property
    def height(self) -> int:
        return self._size.y

    @property
    def tile_size(self):
        return self._tile_size

    @property
    def tile_width(self):
        return self._tile_size.x

    @property
    def tile_height(self):
        return self._tile_size.y



from typing import Union, Tuple

from sfml import sf

from ..data.game_obj import GameObject
from ..data.callbacks import HasCallbacks, callback
from ..core.sprites import AnimPlayer
from ..core.text import BitmapText
from ..core.utils import to_Vector2
from .styles import ButtonStyle


class Button(GameObject, HasCallbacks, sf.Drawable):
    def __init__(self, text: str, style: ButtonStyle = ButtonStyle(),
                 position: Union[Tuple[float, float], sf.Vector2] = (0, 0)):
        super().__init__()
        self._text = text
        self._position = to_Vector2(position)
        self._style = style
        self._render_texture = sf.RenderTexture(self._style.width, self._style.height)
        self.anim_player = AnimPlayer(self._style.anim)
        self._hovered = False

        self._create()

    @property
    def style(self):
        return self._style

    @property
    def size(self):
        return self._style.size

    @property
    def position(self):
        return self._bg.position

    @position.setter
    def position(self, value):
        self._position = to_Vector2(value)
        self._bg.position = self._position
        self._front.position = self._position + self._style.size / 2 \
                               - sf.Vector2(self._front.global_bounds.width / 2, self._front.global_bounds.height / 2)

    @property
    def hovered(self):
        return self._hovered

    def _create(self):
        self.anim_player.play(self._style.anim)
        if isinstance(self._style.background, sf.Color):
            self._bg = sf.RectangleShape((self._style.width, self._style.height))
            self._bg.fill_color = self._style.background
        else:
            self._bg = sf.Sprite(self._style.background)
            if self._style.anim is not None:
                self._bg.texture_rectangle = self._style.anim.frames[0].rectangle
        self._bg.position = self._position

        if isinstance(self._style.font, sf.Font):
            self._front = sf.Text(self._text)
            self._front.font = self._style.font
            self._front.character_size = self._style.font_size
        else:
            self._front = BitmapText(self._text, self._style.font)
            self._front.font = self._style.font
        self._front.position = self._position + sf.Vector2(self.style.width, self.style.height) / 2\
                               - sf.Vector2(self._front.global_bounds.width / 2, self._front.global_bounds.height / 2) \
                               - sf.Vector2(self._front.global_bounds.left, self._front.global_bounds.top)

    @callback("on_press")
    def on_press(self, user_fn):
        return user_fn

    def press(self):
        self.callbacks.call("on_press")

    @callback("on_hover")
    def on_hover(self, user_fn):
        return user_fn

    def hover(self):
        self._hovered = True
        self.callbacks.call("on_hover")

    @callback("on_unhover")
    def on_unhover(self, user_fn):
        return user_fn

    def unhover(self):
        self._hovered = False
        self.callbacks.call("on_unhover")

    def update(self):
        if isinstance(self._bg, sf.Sprite):
            self._bg.texture_rectangle = self.anim_player.active_frame.rectangle
            self.anim_player.update()
            self._bg.origin = self.anim_player.active_frame.origin

    def draw(self, target, states):
        target.draw(self._bg, states)
        target.draw(self._front, states)



from typing import List

from sfml import sf

from ..data.game_obj import GameObject
from ..data.keys import Keyboard
from ..data.callbacks import callback, HasCallbacks
from ..core.utils import to_Vector2
from .buttons import Button
from .styles import MenuStyle, BoxBorder


class Menu(GameObject, HasCallbacks, sf.Drawable):
    def __init__(self, name: str, style: MenuStyle, *args: Button):
        super().__init__()
        self.name = name
        self._style = style
        self._buttons: List[Button] = []
        self.cursor_index = 0
        self.opened = False
        if isinstance(self._style.background, sf.Color):
            self._bg = sf.RectangleShape((self._style.width, self._style.height))
            self._bg.fill_color = self._style.background
        elif isinstance(self._style.background, sf.Texture):
            self._bg = sf.Sprite(self._style.background)
        elif isinstance(self._style.background, BoxBorder):
            texture = self._style.background.generate_texture(self._style.size.x, self._style.size.y)
            self._bg = sf.Sprite(texture)

        self._bg.position = (0, 0)
        for btn in args:
            self.add_button(btn)

    @property
    def style(self):
        return self._style

    @property
    def position(self):
        return self._bg.position

    @position.setter
    def position(self, value):
        self._bg.position = to_Vector2(value)

    @property
    def center(self):
        return self.position + self._style.size / 2

    @center.setter
    def center(self, value):
        self.position = to_Vector2(value) - self._style.size / 2

    @property
    def buttons(self):
        return self._buttons

    def add_button(self, button: Button):
        if isinstance(button, Button):
            self._buttons.append(button)
            button.anim_player.stop()
        else:
            raise TypeError("Menu button should be a Button instance.")

    @callback('on_open')
    def on_open(self, user_fn):
        return user_fn

    def open(self):
        for btn in self._buttons:
            if btn == self._buttons[self.cursor_index]:
                btn.anim_player.resume()
            else:
                btn.anim_player.stop()
        self.game.menus.append(self)
        self.opened = True
        self.callbacks.call("on_open")

    @callback('on_close')
    def on_close(self, user_fn):
        return user_fn

    def close(self):
        self.game.menus.remove(self)
        self.opened = False
        self.callbacks.call("on_close")

    def event_handler(self, event: sf.Event):
        if event == sf.Event.KEY_RELEASED:
            if event['code'] == Keyboard.SPACE:
                self._buttons[self.cursor_index].press()
            elif event['code'] == Keyboard.ESCAPE:
                self.close()
            elif event['code'] == Keyboard.DOWN:
                self._buttons[self.cursor_index].unhover()
                self.cursor_index = (self.cursor_index + 1) % len(self._buttons)
                self._buttons[self.cursor_index].hover()
            elif event['code'] == Keyboard.UP:
                self._buttons[self.cursor_index].unhover()
                self.cursor_index = (self.cursor_index - 1) % len(self._buttons)
                self._buttons[self.cursor_index].hover()

    def update(self):
        for btn in self._buttons:
            btn.update()

    def draw(self, target, states):
        target.draw(self._bg, states)
        for btn in self._buttons:
            target.draw(btn, states)


class VerticalMenu(Menu):
    def __init__(self, name: str, style: MenuStyle, *args: Button):
        super().__init__(name, style, *args)

    def open(self):
        pos = self.position.__copy__() + sf.Vector2(self.style.padding[3], self.style.padding[0])
        for i, btn in enumerate(self._buttons):
            pos += sf.Vector2(0, btn.style.margin[0])
            btn.position = pos
            btn.position += sf.Vector2(btn.style.margin[3], 0)
            pos += sf.Vector2(0, btn.style.bounds.y)
        super().open()



from typing import Union, Tuple, Optional

from sfml import sf

from ..data.rect import Rect
from .utils import to_Vector2


class BitmapGlyph:
    def __init__(self, texture_rect, character, spacing):
        self.texture_rectangle = texture_rect
        self.character = character
        self.spacing = spacing


class BitmapFont:
    def __init__(self, texture: sf.Texture, char_size: Union[Tuple[int, int], sf.Vector2], chars_map=None, spacings_map=None):
        self.texture = texture

        self.char_size = to_Vector2(char_size)

        self.chars_map = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"
        if chars_map:
            self.chars_map = chars_map

        self.spacings_map = {}
        if spacings_map:
            self.spacings_map = spacings_map

        self.glyphs = {}
        i = 0
        for y in range(0, self.texture.height, self.char_size.y):
            for x in range(0, self.texture.width, self.char_size.x):
                if i < len(self.chars_map):
                    character = self.chars_map[i]
                    spacing = self.char_size.x
                    if character in self.spacings_map:
                        spacing = self.spacings_map[character]
                    self.glyphs[character] = BitmapGlyph(Rect((x, y), self.char_size), character, spacing)
                i += 1

    def get_glyph(self, character: str) -> BitmapGlyph:
        return self.glyphs[character]

    def get_glyph_sprite(self, character: str) -> sf.Sprite:
        spr = sf.Sprite(self.texture)
        spr.texture_rectangle = self.glyphs[character].texture_rectangle
        return spr


class BitmapText(sf.Drawable):
    def __init__(self, text: str, font: BitmapFont = None):
        super().__init__()
        self._font: Optional[BitmapFont] = None
        self._text = text
        self._render_texture = sf.RenderTexture(1, 1)
        self._sprite = sf.Sprite(self._render_texture.texture)
        self._glyphs_sprites = []
        if font is not None:
            self.font = font
        self.position = (0, 0)

    @property
    def text(self):
        return self._text

    @property
    def font(self):
        return self._font

    @font.setter
    def font(self, value):
        if isinstance(value, BitmapFont):
            self._font = value
            x = 0
            self._glyphs_sprites = []
            for character in self.text:
                glyph = self._font.get_glyph(character)
                glyph_spr = self._font.get_glyph_sprite(character)
                glyph_spr.position = (x, 0)
                self._glyphs_sprites.append(glyph_spr)
                x += glyph.spacing
            self.update()
        else:
            raise TypeError("BitmapText font argument should be a BitmapFont instance")

    @property
    def position(self) -> sf.Vector2:
        return self._sprite.position

    @position.setter
    def position(self, value):
        self._sprite.position = to_Vector2(value)

    @property
    def origin(self):
        return self._sprite.origin

    @property
    def width(self):
        res = 0
        for c in self.text:
            res += self._font.get_glyph(c).spacing
        return res

    @property
    def height(self):
        return self._font.char_size.y

    @property
    def global_bounds(self):
        return self._sprite.global_bounds

    def update(self):
        self._render_texture = sf.RenderTexture(self.width, self.height)
        self._render_texture.clear(sf.Color.TRANSPARENT)
        for spr in self._glyphs_sprites:
            self._render_texture.draw(spr)
        self._render_texture.display()
        self._sprite.texture = self._render_texture.texture

    def draw(self, target, states):
        if self.font:
            target.draw(self._sprite, states)



from typing import Union, Tuple

from sfml import sf


def to_Vector2(arg: Union[Tuple[int, int], sf.Vector2]) -> sf.Vector2:
    if isinstance(arg, sf.Vector2):
        return arg
    else:
        return sf.Vector2(arg[0], arg[1])


from typing import Union, Tuple, List

from sfml import sf

from ..data.rect import Rect
from .utils import to_Vector2


class AnimFrame:
    def __init__(self, rectangle: Rect, duration: int, origin: Union[Tuple[int, int], sf.Vector2] = (0, 0)):
        self.rectangle: Rect = rectangle    # texture rectangle of the frame
        self.duration: int = duration       # duration of the frame in milliseconds
        self.origin: Union[Tuple[int, int], sf.Vector2] = to_Vector2(origin)    # origin of the frame


class Anim:
    """Anim is used to describe an animation

    For each frame, it contains data about :

        - texture rectangle       (rect.Rect or sf.Rect instance)
        - duration for each frame (list of int)
        - origin of the frame     ((0,0) by default )

    You can also specify if the animation should loop or not (True by default)
    """
    def __init__(self, frames: List[AnimFrame], loop: bool = True):
        self.frames = frames
        self.frames_count = len(self.frames)
        self.loop = loop

    def add(self, frame: AnimFrame):
        self.frames.append(frame)
        self.frames_count += 1


class AnimPlayer:
    def __init__(self, animation: Anim):
        self.anim = animation
        self.active_frame = self.anim.frames[0]
        self.index = 0
        self.playing = True
        self.clock = sf.Clock()

    def play(self, animation: Anim):
        if animation != self.anim:
            self.anim = animation
            self.index = 0
            self.playing = True
            self.clock.restart()

    def stop(self):
        self.playing = False
        self.index = 0
        self.active_frame = self.anim.frames[0]

    def pause(self):
        self.playing = False

    def resume(self):
        self.playing = True

    def update(self):
        if self.playing:
            if self.clock.elapsed_time.milliseconds >= self.anim.frames[self.index].duration:
                self.index += 1
                if self.index >= self.anim.frames_count:
                    if self.anim.loop:
                        self.index = 0
                    else:
                        self.playing = False
                        self.index -= 1
                self.clock.restart()
            self.active_frame = self.anim.frames[self.index]


class Sprite:
    def __init__(self, name: str, texture, anims: dict):
        self.name = name
        self.texture = texture
        self.anims = anims



from typing import Dict

from sfml import sf

from ..data.game_obj import GameObject
from ..data import rect, keys
from .sprites import Anim, AnimPlayer


class BaseEntity(GameObject, sf.Drawable):
    def __init__(self, name: str, data):
        super().__init__()
        self.data = data
        self.name = name
        self._anim_state = "idle"

        self.anims: Dict[str, Anim] = self.data.anims
        self.anim_player: AnimPlayer = AnimPlayer(self.anims[self._anim_state])
        self.animate = False if self.anims is None else True

        self.sprite = sf.Sprite(self.data.texture)
        self.sprite.texture_rectangle = self.anims[self.anim_state].frames[0].rectangle
        self.sprite.origin = self.anims[self.anim_state].frames[0].origin
        self.direction = sf.Vector2(1, 1)

        self.gx = self.sprite.position.x
        self.gy = self.sprite.position.y

        self.rx = 0
        self.ry = 0

        self.collision_box = self.global_bounds
        self.collision_box_shape = sf.RectangleShape((self.collision_box.width, self.collision_box.height))
        self.collision_box_shape.position = (self.collision_box.left, self.collision_box.top)
        self.collision_box_shape.fill_color = sf.Color(200, 0, 0, 150)


    @property
    def anim_state(self):
        return self._anim_state

    @anim_state.setter
    def anim_state(self, value: str):
        if self._anim_state != value:
            if value in self.anims:
                self._anim_state = value
                self.anim_player.play(self.anims[self._anim_state])
                self.sprite.texture_rectangle = self.anim_player.active_frame.rectangle
                self.sprite.origin = self.anim_player.active_frame.origin
            else:
                raise KeyError(f"Entity named {self.name} has no animation anim_state named {value}")

    @property
    def position(self):
        return sf.Vector2(self.x, self.y)

    @position.setter
    def position(self, value: tuple):
        if isinstance(value, sf.Vector2):
            self.x = value.x
            self.y = value.y
        else:
            self.x = value[0]
            self.y = value[1]

    @property
    def x(self):
        return (self.gx + self.rx) * 16

    @x.setter
    def x(self, value):
        self.gx = value // 16
        self.rx = (value - self.gx*16)/16

    @property
    def y(self):
        return (self.gy + self.ry) * 16

    @y.setter
    def y(self, value):
        self.gy = value // 16
        self.ry = (value - self.gy*16)/16

    @property
    def global_bounds(self):
        bounds = self.sprite.global_bounds
        x, y, width, height = bounds.left, bounds.top, bounds.width, bounds.height
        return rect.Rect((x, y), (width, height))

    def update(self, dt: float, keys: list = None):
        self.sprite.position = (self.x, self.y)
        self.sprite.position = (round(self.x), round(self.y))
        self.sprite.ratio = self.direction
        self.collision_box = self.global_bounds
        self.collision_box_shape.size = (self.global_bounds.width, self.global_bounds.height)
        self.collision_box_shape.position = (self.global_bounds.left, self.global_bounds.top)

        self.update_anim()

    def update_anim(self):
        self.anim_player.update()
        if self.sprite.texture_rectangle != self.anim_player.active_frame.rectangle:
            self.sprite.texture_rectangle = self.anim_player.active_frame.rectangle
            self.sprite.origin = self.anim_player.active_frame.origin

    def draw(self, target, anim_states):
        target.draw(self.sprite, anim_states)
        if self.game.debug:
            target.draw(self.collision_box_shape)


class PlatformerEntity(BaseEntity):
    def __init__(self, name: str, data):
        super().__init__(name, data)
        self.sprite.origin = self.anims[self.anim_state].frames[0].origin

        self.controls = {
            'right': keys.Keyboard.RIGHT,
            'left': keys.Keyboard.LEFT,
            'up': keys.Keyboard.UP,
            'down': keys.Keyboard.DOWN
        }

        self.velocity = sf.Vector2(0, 0)
        self.jump_velocity = sf.Vector2(0, -18)
        self.gravity = sf.Vector2(0, 1)
        self.onground = False
        self.jumping = False
        self.jump_count = 2
        self.remaining_jumps = 2
        self.falling = True

    def jump(self):
        col_over = []
        if self.onground:
            # these 2 lines get the nearest collision box over the entity
            self.game.level.collisions.sort(key=lambda b:abs(b.bottom-self.collision_box.top))
            col_over = [x for x in self.game.level.collisions if x.bottom < self.collision_box.top and x.left <= self.x <= x.right]

        if not col_over or (col_over and self.collision_box.top // 16 != col_over[0].bottom // 16):
            self.jumping = True
            self.onground = False
            self.velocity.y = self.jump_velocity.y
            self.remaining_jumps -= 1

    def land(self):
        if not self.onground:
            self.jumping = False
            self.onground = True
            self.falling = False
            self.remaining_jumps = self.jump_count

    def idle(self):
        if -0.05 < self.velocity.x < 0.05:
            self.velocity.x = 0
        else:
            if self.onground:
                self.velocity.x = self.velocity.x * 0.80
            else:
                if self.velocity.x < 0:
                    self.velocity.x += 0.3
                else:
                    self.velocity.x -= 0.3
        self.anim_state = "idle"

    def walk_right(self):
        self.velocity.x = min(self.velocity.x + 0.5, 12)
        self.direction.x = 1
        self.anim_state = "walk"

    def walk_left(self):
        self.velocity.x = max(self.velocity.x - 0.5, -12)
        self.direction.x = -1
        self.anim_state = "walk"

    def update(self, dt: float, keys: list = None):
        if not self.onground:
            self.velocity += self.gravity
        if self.velocity.y > 0 and not self.onground:
            self.falling = True
        else:
            self.falling = False
        self.rx += self.velocity.x * dt

        if self.game.level:
            for box in self.game.level.collisions:
                if box.top <= self.collision_box.top < box.bottom or box.top < self.collision_box.bottom <= box.bottom \
                        or self.collision_box.top <= box.top < self.collision_box.bottom or self.collision_box.top < box.bottom <= self.collision_box.bottom:
                    # left collision
                    if self.velocity.x < 0:
                        if self.collision_box.left//16 == (box.right -1) // 16 and self.rx < 0.3:
                            self.rx -= self.velocity.x*dt
                            self.velocity.x = 0
                    # rigth collision
                    if self.velocity.x > 0:
                        if self.collision_box.right//16 == box.left // 16 and self.rx > 0.7:
                            self.rx -= self.velocity.x*dt
                            self.velocity.x = 0

        while self.rx > 1: self.rx -= 1; self.gx += 1
        while self.rx < 0: self.rx += 1; self.gx -= 1

        self.ry += self.velocity.y * dt
        bot_collision = False

        if self.game.level:
            for box in self.game.level.collisions:
                if box.left / 16 <= self.gx < (box.right) / 16:
                    # top collision
                    if self.velocity.y < 0:
                        if self.collision_box.top//16 == (box.bottom - 1) // 16 and self.ry < 0.99:
                            self.y = box.bottom + self.sprite.origin.y
                            self.falling = True
                            self.velocity.y = 0
                    # bottom collision
                    if self.gy + 1 == box.top / 16 and self.ry >= 0.99:
                        self.ry = 0.99
                        self.velocity.y = 0
                        self.land()
                        bot_collision = True
                        break

        if not bot_collision:
            self.onground = False
        while self.ry > 1: self.ry -= 1; self.gy += 1
        while self.ry < 0: self.ry += 1; self.gy -= 1

        super().update(dt)


from .app import App
from .camera import Camera
from .scenes import Scene
from .layers import Layer, Mask
from .sprites import Sprite, AnimFrame, Anim
from .entities import BaseEntity, PlatformerEntity
from .text import BitmapText, BitmapFont
from .debug import Logger
from .window import RenderWindow
from . import transitions

__all__ = [
    'App', 'Camera', 'Scene', 'Layer', 'Mask',
    'Sprite', 'Anim', 'AnimFrame', 'BitmapText', 'BitmapFont',
    'BaseEntity', 'PlatformerEntity',
    'transitions',
    'Logger', 'RenderWindow'
]



from .core import *
from .data import *
from .reslib import *
from .tilemapping import *
from . import ui

__all__ = [
    'App', 'Camera', 'Scene', 'Layer', 'Mask', 'Sprite', 'Anim', 'AnimFrame', 'BitmapText', 'BitmapFont',
    'BaseEntity', 'PlatformerEntity',
    'transitions',

    'Keyboard', 'Rect', 'SCREEN_W', 'SCREEN_H', 'LineShape', 'EllipseShape',

    'Res', 'find_resource', 'RES_DIR',

    'TiledMap',

    'ui'
]

__author__ = "Modar NASSER"
__version__ = '0.1.3'



from sfml import sf

desktop_mode = sf.VideoMode.get_desktop_mode()
all_modes = sf.VideoMode.get_fullscreen_modes()

fullscreen_mode = all_modes[0]
SCREEN_W = float(fullscreen_mode.width)
SCREEN_H = float(fullscreen_mode.height)



from typing import Union, Tuple

from sfml import sf


class Rect(sf.Rect):
    def __init__(self, pos: Union[Tuple[float, float], sf.Vector2], size: Union[Tuple[float, float], sf.Vector2]):
        super().__init__(pos, size)

    @classmethod
    def from_rect(cls, sf_rect: sf.Rect):
        return Rect((sf_rect.left, sf_rect.top), (sf_rect.width, sf_rect.height))

    @property
    def size(self):
        return sf.Vector2(self.width, self.height)

    @property
    def right(self):
        return self.left + self.width

    @property
    def bottom(self):
        return self.top + self.height

    @property
    def topleft(self):
        return sf.Vector2(self.left, self.top)

    @property
    def bottomleft(self):
        return sf.Vector2(self.left, self.bottom)

    @property
    def topright(self):
        return sf.Vector2(self.right, self.top)

    @property
    def bottomright(self):
        return sf.Vector2(self.right, self.bottom)

    @property
    def center(self):
        return sf.Vector2(self.left + self.width/2, self.top + self.height/2)

    @property
    def topcenter(self):
        return sf.Vector2(self.center.x, self.top)

    def intersects(self, other_rect):
        if isinstance(other_rect, Rect):
            if self.left >= other_rect.right or \
                    self.right <= other_rect.left or \
                    self.top >= other_rect.bottom or \
                    self.bottom <= other_rect.top:
                return False
            return True
        elif isinstance(other_rect, sf.Rect):
            if self.left >= other_rect.left + other_rect.width or \
                    self.left + self.width <= other_rect.left or \
                    self.top >= other_rect.top + other_rect.height or \
                    self.top + self.height <= other_rect.top:
                return False
            return True



from typing import Tuple, Union

from sfml import sf

from ..data.rect import Rect
from ..core.text import BitmapFont
from ..core.sprites import Anim
from ..core.utils import to_Vector2


class BoxBorder():
    def __init__(self, texture: sf.Texture, tex_pos: Union[Tuple[int, int], sf.Vector2], tile_size: Union[Tuple[int, int], sf.Vector2]):
        self._texture = texture
        self._pattern_pos = to_Vector2(tex_pos)
        self._tile_size = to_Vector2(tile_size)
        self._render_textures = {}

        top_left = sf.Sprite(self._texture)
        top_left.texture_rectangle = Rect(self._pattern_pos, self._tile_size)

        top_right = sf.Sprite(self._texture)
        top_right.texture_rectangle = Rect(self._pattern_pos + sf.Vector2(2*self._tile_size.x, 0), self._tile_size)

        bot_left = sf.Sprite(self._texture)
        bot_left.texture_rectangle = Rect(self._pattern_pos + sf.Vector2(0, 2*self._tile_size.y), self._tile_size)

        bot_right = sf.Sprite(self._texture)
        bot_right.texture_rectangle = Rect(self._pattern_pos + self._tile_size + self._tile_size, self._tile_size)

        self.corners = {
            'top_left': top_left,
            'top_right': top_right,
            'bot_left': bot_left,
            'bot_right': bot_right
        }

        self.vertex_pos = [
            sf.Vector2(0, 0),
            sf.Vector2(self._tile_size.x, 0),
            sf.Vector2(self._tile_size.x, self._tile_size.y),
            sf.Vector2(0, self._tile_size.y),
        ]

        self.tex_coords = {
            'top': [
                self._pattern_pos + sf.Vector2(self._tile_size.x, 0),
                self._pattern_pos + sf.Vector2(self._tile_size.x, 0) + sf.Vector2(self._tile_size.x, 0),
                self._pattern_pos + sf.Vector2(self._tile_size.x, 0) + self._tile_size,
                self._pattern_pos + sf.Vector2(self._tile_size.x, 0) + sf.Vector2(0, self._tile_size.y),
            ],
            'right': [
                self._pattern_pos + sf.Vector2(2 * self._tile_size.x, self._tile_size.y),
                self._pattern_pos + sf.Vector2(2 * self._tile_size.x, self._tile_size.y) + sf.Vector2(self._tile_size.x, 0),
                self._pattern_pos + sf.Vector2(2 * self._tile_size.x, self._tile_size.y) + self._tile_size,
                self._pattern_pos + sf.Vector2(2 * self._tile_size.x, self._tile_size.y) + sf.Vector2(0, self._tile_size.y),
            ],
            'bot': [
                self._pattern_pos + sf.Vector2(self._tile_size.x, 2 * self._tile_size.y),
                self._pattern_pos + sf.Vector2(self._tile_size.x, 2 * self._tile_size.y) + sf.Vector2(self._tile_size.x, 0),
                self._pattern_pos + sf.Vector2(self._tile_size.x, 2 * self._tile_size.y) + self._tile_size,
                self._pattern_pos + sf.Vector2(self._tile_size.x, 2 * self._tile_size.y) + sf.Vector2(0, self._tile_size.y),
            ],
            'left': [
                self._pattern_pos + sf.Vector2(0, self._tile_size.y),
                self._pattern_pos + sf.Vector2(0, self._tile_size.y) + sf.Vector2(self._tile_size.x, 0),
                self._pattern_pos + sf.Vector2(0, self._tile_size.y) + self._tile_size,
                self._pattern_pos + sf.Vector2(0, self._tile_size.y) + sf.Vector2(0, self._tile_size.y),
            ],
            'center': [
                self._pattern_pos + self._tile_size,
                self._pattern_pos + self._tile_size + sf.Vector2(self._tile_size.x, 0),
                self._pattern_pos + self._tile_size + self._tile_size,
                self._pattern_pos + self._tile_size + sf.Vector2(0, self._tile_size.y),
            ]
        }

    def generate_texture(self, width: int, height: int) -> sf.Texture:
        w = width//self._tile_size.x * self._tile_size.x
        h = height//self._tile_size.y * self._tile_size.y
        if (w, h) in self._render_textures: return self._render_textures[(w, h)].texture

        self.corners['top_left'].position = (0, 0)
        self.corners['top_right'].position = (w - self._tile_size.x, 0)
        self.corners['bot_left'].position = (0, h - self._tile_size.y)
        self.corners['bot_right'].position = (w - self._tile_size.x, h - self._tile_size.y)

        horizontal_size = ((w - 2 * self._tile_size.x) // self._tile_size.x) * 4
        vertical_size = ((h - 2 * self._tile_size.y) // self._tile_size.y) * 4
        fill_size = (((w - 2 * self._tile_size.x) // self._tile_size.x) * (h - 2 * self._tile_size.y) // self._tile_size.y) * 4
        sizes = {'top': horizontal_size, 'right': vertical_size, 'bot': horizontal_size, 'left': vertical_size, 'center': fill_size}
        offset = {
            'top': sf.Vector2(self._tile_size.x, 0),
            'right': sf.Vector2(w - self._tile_size.x, self._tile_size.y),
            'bot': sf.Vector2(self._tile_size.x, h - self._tile_size.y),
            'left': sf.Vector2(0, self._tile_size.y),
            'center': self._tile_size,
        }
        increment = {
            'top': sf.Vector2(self._tile_size.x/4, 0),
            'right': sf.Vector2(0, self._tile_size.y/4),
            'bot': sf.Vector2(self._tile_size.x/4, 0),
            'left': sf.Vector2(0, self._tile_size.y/4),
            'center': sf.Vector2(self._tile_size.x/4, self._tile_size.y/4)
        }

        vertices = []

        for side in self.tex_coords:
            vert = sf.VertexArray(sf.PrimitiveType.QUADS)
            vert.resize(sizes[side])
            for i in range(0, sizes[side], 4):
                for j in range(4):
                    vert[i + j].tex_coords = self.tex_coords[side][j]
                    if side == 'center':
                        center_offset = sf.Vector2(
                            (increment[side].x * i) % (w - 2 * self._tile_size.x),
                            (increment[side].y * i) // (w - 2 * self._tile_size.x) * self._tile_size.y
                        )
                        vert[i + j].position = self.vertex_pos[j] + offset[side] + center_offset
                    else:
                        vert[i + j].position = self.vertex_pos[j] + offset[side] + increment[side] * i
            vertices.append(vert)

        render_texture = sf.RenderTexture(w, h)
        render_texture.clear(sf.Color.TRANSPARENT)

        render_texture.draw(self.corners['top_left'])
        render_texture.draw(self.corners['top_right'])
        render_texture.draw(self.corners['bot_left'])
        render_texture.draw(self.corners['bot_right'])

        for vert in vertices:
            render_texture.draw(vert, sf.RenderStates(texture=self._texture))
        render_texture.display()

        self._render_textures[(w, h)] = render_texture
        return render_texture.texture


class ButtonStyle:
    __slots__ = ["width", "height", "size", "bounds", "margin", "font",
                 "color", "font_size", "background", "anim"]

    def __init__(self,
                 width: int = 0, height: int = 0,
                 margin: Tuple[int, int, int, int] = (0, 0, 0, 0),
                 font: Union[sf.Font, BitmapFont] = None,
                 color: sf.Color = sf.Color.WHITE,
                 font_size: int = 16,
                 background: Union[sf.Color, sf.Texture] = sf.Color.TRANSPARENT,
                 anim: Anim = None,
                 ):
        self.width = width
        self.height = height
        self.size = sf.Vector2(width, height)
        self.bounds = sf.Vector2(margin[3] + width + margin[1], margin[0] + height + margin[2])
        self.margin = margin
        self.font = font
        self.color = color
        self.font_size = font_size
        self.background = background
        self.anim = anim


class MenuStyle:
    __slots__ = ["width", "height", "size", "bounds", "padding", "background"]

    def __init__(self,
                 width: int = 0, height: int = 0,
                 padding: Tuple[int, int, int, int] = (0, 0, 0, 0),
                 background: Union[sf.Color, sf.Texture, BoxBorder] = sf.Color.TRANSPARENT
                 ):
        self.size = sf.Vector2(padding[3] + width + padding[1] , padding[0] + height  + padding[3])
        self.width = self.size.x
        self.height = self.size.y
        self.padding = padding
        self.background = background




from .resource_loader import load_resources, Dir
from .resource_path import find_resource, RES_DIR
from .tileset_manager import TilesetManager


class ResMeta(type):
    """
    To use the res manager, there are some naming convention you should follow:
        - your files and folders should respect python variable naming rules
        - your files and folders names should not start with an underscore
        - " Textures " , " Maps " and " Fonts " are reserved. Avoid using those as a name.
    """
    _ready: bool = False
    _assets: Dir = Dir("assets")
    _textures: Dir = Dir("textures")
    _maps: Dir = Dir("maps")
    _fonts: Dir = Dir("fonts")
    _parent = None

    @property
    def Textures(cls) -> Dir:
        return cls._textures

    @property
    def Maps(cls) -> Dir:
        return cls._maps

    @property
    def Fonts(cls) -> Dir:
        return cls._fonts

    @property
    def is_ready(cls) -> bool:
        return cls._ready

    @classmethod
    def __iter__(mcs):
        for x in mcs._assets:
            yield x

    @classmethod
    def values(mcs):
        return mcs._assets.values()

    @classmethod
    def items(mcs):
        return mcs._assets.items()

    def __getattr__(cls, item) -> Dir:
        if item[0] == "_":
            return cls.__dict__[item]
        return getattr(cls._assets, item)

    def __setattr__(cls, key, value):
        if key[0] != "_":
            raise AttributeError("You can not set attribute for Res class.")
        else:
            super().__setattr__(key, value)


class Res(metaclass=ResMeta):
    """
    A Resource manager. Automatically loads all resources found in `assets` directory
    To access a resource, you can use : `Res.folder_name.resource_name`
    Res loads all assets recursively, so you can access nested directories this way :
        `Res.folder1.folder2.folder3.resource_name`
    Res.Textures contains all textures at the same level.
    Res.Maps contains all tiled maps.
    Res.Fonts contains all fonts.
    """
    @classmethod
    def load(cls):
        """Load all resources into Res.
        You need to call this method once at the start of your game
        if you want to use the resource manager.
        """
        cls._assets._parent = cls
        load_resources(cls._assets, find_resource(RES_DIR))

        for tileset in TilesetManager:
            tileset.load()

        for tile_map in cls.Maps.values():
            tile_map.load()

        cls._ready = True

    @classmethod
    def get_tree(cls):
        return cls._assets

    @classmethod
    def all_files(cls):
        """Returns all the loaded resources without any particular order"""
        return cls._assets.all_files()

    @classmethod
    def print_tree(cls):
        """Prints the tree of the assets folder """
        print(f"\n{RES_DIR}")
        cls._assets.print_tree(indent=1)
        print("")



from typing import Union, Optional, Tuple
import random

from sfml import sf

from ..data.game_obj import GameObject
from ..data.rect import Rect
from .utils import to_Vector2


class Camera(GameObject, sf.View):
    def __init__(self, name, render_order):
        super().__init__()
        self.name = name
        self.render_order = render_order
        self.reference = None
        self.state = CameraState()
        self.frames_delay = 15
        self.base_pos = sf.Vector2(0, 0)
        self.base_size = sf.Vector2(0, 0)
        self.vp_base_pos = sf.Vector2(0, 0)
        self.vp_base_size = sf.Vector2(1, 1)
        self.offset = sf.Vector2(0, 0)
        self.visible: bool = True
        self._scene = []

    def reset(self, position: Union[sf.Vector2, Tuple[float, float]], size: Union[sf.Vector2, Tuple[float, float]]):
        if type(position) in [sf.Vector2, tuple] and type(size) in [sf.Vector2, tuple]:
            super().reset(sf.Rect(position, size))
            self.base_pos = to_Vector2(position)
            self.base_size = to_Vector2(size)
        else:
            raise TypeError("position and size arguments of Camera.reset() should be a sf.Vector or tuple.")

    def reset_viewport(self, position: Union[sf.Vector2, Tuple[float, float]], size: Union[sf.Vector2, Tuple[float, float]]):
        if type(position) in [sf.Vector2, tuple] and type(size) in [sf.Vector2, tuple]:
            self.viewport = sf.Rect(position, size)
            self.vp_base_pos = to_Vector2(position)
            self.vp_base_size = to_Vector2(size)
        else:
            raise TypeError("position and size arguments of Camera.reset_viewport() should be a sf.Vector or tuple.")

    @property
    def scene(self):
        return self._scene[0]

    @scene.setter
    def scene(self, val):
        if not self._scene:
            self._scene.append(val)
        else:
            self._scene[0] = val

    def has_scene(self):
        if self._scene:
            return True
        return False

    @property
    def position(self):
        return sf.Vector2(self.left, self.top)

    @property
    def size(self) -> sf.Vector2:
        return sf.Vector2(self.base_size.x, self.base_size.y)

    @property
    def left(self) -> float:
        return self.center.x - self.size.x / 2

    @left.setter
    def left(self, value: int):
        self.center = (value + self.size.x / 2, self.center.y)

    @property
    def top(self) -> float:
        return self.center.y - self.size.y / 2

    @top.setter
    def top(self, value: int):
        self.center = (self.center.x, value + self.size.y/2)

    @property
    def right(self) -> float:
        return self.center.x + self.size.x / 2

    @right.setter
    def right(self, value: int):
        self.center = (value - self.size.x/2, self.center.y)

    @property
    def bottom(self) -> float:
        return self.center.y + self.size.y / 2

    @bottom.setter
    def bottom(self, value: int):
        self.center = (self.center.x, value - self.size.y / 2)

    @property
    def bounds(self) -> Rect:
        return Rect((self.left, self.top), (self.size.x, self.size.y))

    def follow(self, entity):
        self.reference = entity

    def quake(self, duration: float, amplitude: int, horizontal: bool = True, vertical: bool = True):
        self.state = QuakeState(duration, amplitude, horizontal, vertical)

    def update(self):
        self.move(- self.offset.x, - self.offset.y)

        if self.reference:
            dif = self.reference.position - self.center - self.offset
            if self.frames_delay > 0:
                self.move(dif.x/self.frames_delay, dif.y/self.frames_delay)
                if self.left < 0:
                    self.left = 0
                elif self.right > self.game.level.width*16:
                    self.right = self.game.level.width*16

                if self.top < 0:
                    self.top = 0
                elif self.bottom > self.game.level.height*16:
                    self.bottom = self.game.level.height*16
            else:
                self.center = self.reference.position

        self.offset = self.state.offset
        self.move(self.offset.x, self.offset.y)

        if self.state.expired:
            self.state = CameraState()


class CameraState:
    def __init__(self):
        self.name = "idle"
        self.timer = sf.Clock()
        self.duration: Optional[float] = None

    @property
    def offset(self) -> sf.Vector2:
        return sf.Vector2(0, 0)

    @property
    def expired(self) -> bool:
        if not self.duration:
            return False
        return self.timer.elapsed_time.seconds >= self.duration


class QuakeState(CameraState):
    def __init__(self, duration: float, amplitude: int, horizontal: bool = True, vertical: bool = True):
        super().__init__()
        self.name = "quake"
        self.duration = duration
        self.amplitude = amplitude
        self.direction = sf.Vector2(0, 0)
        self.direction.x = 1 if horizontal else 0
        self.direction.y = 1 if vertical else 0

    @property
    def offset(self) -> sf.Vector2:
        ratio = 1 - self.timer.elapsed_time.seconds/self.duration
        offx = random.randrange(-self.amplitude, self.amplitude, 1)
        offy = random.randrange(-self.amplitude, self.amplitude, 1)
        return sf.Vector2(offx*self.direction.x*ratio, offy*self.direction.y*ratio)



from typing import List, Optional

from sfml import sf

from ..data.game_obj import GameObject
from ..data import const
from ..data.rect import Rect
from .. import ui

from . import camera
from . import scenes
from . import debug
from . import transitions
from . import window


class App:
    def __init__(self, title: str = "NasNas game", w_width: int = 960, w_height: int = 540,
                 v_width: float = 960, v_height: float = 540, fps: Optional[int] = 60):
        """
        Initializes the engine and creates:
            - a window
            - a main scene
            - a camera named game_camera
        Args:
            title (str): title of the game window
            w_width (int): window width
            w_height (int): window height
            v_width (int): game view width
            v_height (int): game view height
            fps (int): desired fps
        """
        GameObject.game = self

        self.W_WIDTH, self.W_HEIGHT = w_width, w_height
        self.V_WIDTH, self.V_HEIGHT = v_width, v_height

        # Game name, displayed in window title
        self.name = title

        self.desired_fps = fps
        self._window = window.RenderWindow(sf.VideoMode(w_width, w_height), self.name)

        if self.desired_fps:
            self._window.framerate_limit = self.desired_fps

        self._level = None

        self._fullscreen = False

        # inputs from keyboard are stored in this list
        self._inputs: List[int] = []
        # list of all cameras/view used in the game
        self._cameras: List[camera.Camera] = []
        # list of all scenes used in the game, usually 1 is enough
        self._scenes: List[scenes.Scene] = []
        # list of all playing transitions
        self._transitions: List[transitions.Transition] = []
        # list of all menus
        self._menus: List[ui.Menu] = []

        # Scene is where everything is drawn on
        self.scene = self.create_scene(w_width, w_height)

        # game camera, usually looks at the main scene (world map, scrolling level ...)
        self.game_camera = self.create_camera("game", 0, Rect((0, 0), (v_width, v_height)))

        # clock and dt used for FPS calculation
        self.clock = sf.Clock()
        self.dt = 1.0

        self.debug = False
        self.debug_texts = []  # list of DebugText

        self.scale_view()

    @property
    def window(self) -> window.RenderWindow:
        return self._window

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, value):
        self._level = value

    @property
    def fullscreen(self) -> bool:
        return self._fullscreen

    @property
    def inputs(self) -> List[int]:
        return self._inputs

    @property
    def cameras(self) -> List[camera.Camera]:
        return self._cameras

    @property
    def scenes(self) -> List[scenes.Scene]:
        return self._scenes

    @property
    def transitions(self) -> List[transitions.Transition]:
        return self._transitions

    @property
    def menus(self) -> List[ui.Menu]:
        return self._menus

    def create_scene(self, width: int, height: int):
        """
        Creates a new Scene and returns it.

        Args:
            width (int): width of the new scene
            height (int): height of the new scene

        Returns:
            Returns the created Scene.
        """
        s = scenes.Scene(width, height)
        self.scenes.append(s)
        return s

    def create_camera(self, name: str, order: int, camwindow: Rect, viewport: Rect = Rect((0, 0), (1, 1))) -> camera.Camera:
        """
        Creates a new Camera and returns it

        Args:
            name (str):
                The name of the new Camera, should be unique.
            order (int):
                The order of display. 0 will be displayed first, 1 will be displayed over 0, etc
            camwindow (Rect):
                The window the Camera is looking at.
                For exemple, a Camera with a window=sf.Rect((0, 10), (20, 20)) will look at a square 20x20 located at (0, 10)
            viewport (Rect):
                The place and surface the Camera will take on the window.
                If we take the previous window, and a viewport of sf.Rect((0.5, 0), (0.2, 0.2)), the window looked at will
                be displayed on a surface equal to 20% x 20% of the window located at (50%, 0) of the window size.

        Returns:
            Returns the created Camera.
        """
        cam = camera.Camera(name, order)
        cam.reset(camwindow.topleft, camwindow.size)
        cam.reset_viewport(viewport.topleft, viewport.size)
        self.cameras.append(cam)
        return cam

    def add_debug_text(self, instance: object, attr_name: str, position: tuple, float_round: int = None):
        """
        Adds a DebugText to self.debug_texts. The text will be drawn at the given position

        Args:
            instance (object):
                An instance of any object.
            attr_name (str):
                Attribute name of the instance.
            position (tuple):
                The position where the text will be drawn
            float_round (int):
                If the attribute value is a float, rounds that float using round(instance.attr_name, float_round)
        """
        if float_round:
            self.debug_texts.append(debug.DebugText(instance, attr_name, position, float_round))
        else:
            self.debug_texts.append(debug.DebugText(instance, attr_name, position))

    def toggle_fullscreen(self):
        if not self.fullscreen:
            self.window.close()
            self._window = window.RenderWindow(sf.VideoMode(const.SCREEN_W, const.SCREEN_H), self.name, sf.Style.NONE)
            self.window.framerate_limit = self.desired_fps
            self.window.mouse_cursor_visible = False
            self.window.vertical_synchronization = True
            self._fullscreen = True
            self._inputs = []
            self.scale_view()
        else:
            self.window.close()
            self._window = window.RenderWindow(sf.VideoMode(self.W_WIDTH, self.W_HEIGHT), self.name, sf.Style.DEFAULT)
            self.window.framerate_limit = self.desired_fps
            self.window.mouse_cursor_visible = True
            self.window.vertical_synchronization = True
            self._fullscreen = False
            self._inputs = []
            self.scale_view()

    def scale_view(self):
        """
        Scales all Cameras to keep a good aspect ratio and avoid deformation when resizing the window.
        """
        # wider than usual window
        if float(self.window.size.x) / float(self.window.size.y) > self.W_WIDTH / self.W_HEIGHT:
            viewport_w = (float(self.window.size.y) * (self.W_WIDTH / self.W_HEIGHT)) / float(self.window.size.x)
            viewport_h = 1
            viewport_x = (1-viewport_w)/2
            viewport_y = 0
        # higher than usual window
        else:
            viewport_w = 1
            viewport_h = (float(self.window.size.x) / (self.W_WIDTH / self.W_HEIGHT)) / float(self.window.size.y)
            viewport_x = 0
            viewport_y = (1-viewport_h)/2

        def new_viewport(x):
            return sf.Rect(
                (viewport_x + x.vp_base_pos.x * (1 - 2 * viewport_x), viewport_y + x.vp_base_pos.y * (1 - 2 * viewport_y)),
                (viewport_w * x.vp_base_size.x, viewport_h * x.vp_base_size.y)
            )
        for cam in self.cameras:
            cam.viewport = new_viewport(cam)
        self.window.ui_view.viewport = new_viewport(self.window.ui_view)

    def _store_inputs(self, event: sf.Event):
        """
        Handles window close event.
        Calls self.scale_view() when the window is resized
        Update inputs list.

        Args:
            event (sf.Event): a window event
        """
        if event == sf.Event.KEY_PRESSED:
            if event['code'] not in self.inputs:
                self.inputs.insert(0, event['code'])
        elif event == sf.Event.KEY_RELEASED:
            if event['code'] in self.inputs:
                self.inputs.remove(event['code'])

    def event_handler(self, event: sf.Event):
        """
        event_handler is called every time there is an event happening (key pressed, mouse move, lost focus ...)
        Override this method to make your own event handling

        Args:
            event (sf.Event): a SFML event
        """
        pass

    def update(self):
        """
        Write here your game logic
        """

    def _render(self):
        """
        Draw all Scenes and Layers in order.
        Layer 0 will be drawn first, layer 1 will be drawn over layer 0 etc
        Internal usage only, do not override or call this method.
        """
        for scene in self.scenes:
            scene.render()
        for cam in self.cameras:
            cam.update()
        self.cameras.sort(key=lambda x: x.render_order)
        for cam in self.cameras:
            if cam.visible:
                self.window.view = cam
                if cam.has_scene():
                    self.window.draw(cam.scene)

        # drawing transitions on top of everything else directly on the window
        self.window.view = self.window.ui_view
        for menu in self.menus:
            menu.update()
            self.window.draw(menu)
        for transition in self.transitions:
            transition.update()
            self.window.draw(transition)

    def run(self):
        """
        Starts the game loop.
        """
        while self.window.is_open:
            self.dt = self.clock.restart().seconds
            self.window.title = self.name+" - FPS:" + str(round(1 / self.dt))

            for event in self.window.events:
                if not self._menus:
                    self._store_inputs(event)
                    self.event_handler(event)
                else:
                    self._inputs = []
                    self._menus[-1].event_handler(event)

            self.update()

            self.window.clear(sf.Color.BLACK)

            self._render()

            self.window.view = self.window.default_view
            if self.debug:
                for txt in self.debug_texts:
                    txt.update()
                    self.window.draw(txt)

            self.window.display()



from functools import wraps


def callback(name):
    def decorator(method):
        @wraps(method)
        def wrapper(self, user_fn):
            self.callbacks.register(name, user_fn)
            return method(self, user_fn)
        return wrapper
    return decorator


class CallbackHandler:
    def __init__(self):
        self._callbacks = {}

    def register(self, name: str, function):
        self._callbacks[name] = function

    def call(self, name: str):
        if name in self._callbacks:
            self._callbacks[name]()


class HasCallbacks:
    def __init__(self, *args, **kwargs):
        # avoid blocking multiple inheritance by forwarding all arguments to super()
        super().__init__(*args, **kwargs)
        self._callbacks = CallbackHandler()

    @property
    def callbacks(self):
        return self._callbacks



from typing import Union, Tuple
import math

from sfml import sf

from ..core.utils import to_Vector2


class LineShape(sf.Drawable):
    def __init__(self, *args: Union[Tuple[float, float], sf.Vector2]):
        super().__init__()
        self._points = []
        self._position = sf.Vector2(0, 0)
        self._vertexarray = sf.VertexArray(sf.PrimitiveType.LINES_STRIP)
        self._color = sf.Color.WHITE
        for arg in args:
            self._points.append(to_Vector2(arg))
        self.update()

    @property
    def point_count(self) -> int:
        return len(self._points)

    @property
    def position(self) -> sf.Vector2:
        return self._position

    @position.setter
    def position(self, pos: Union[Tuple[float, float], sf.Vector2]):
        old_pos = self._position
        self._position = to_Vector2(pos)
        self.update(offset=self._position-old_pos)

    @property
    def color(self) -> sf.Color:
        return self._color

    @color.setter
    def color(self, color: sf.Color):
        self._color = color
        self.update(color=True)

    def add_point(self, point: Union[Tuple[float, float], sf.Vector2]):
        self._points.append(to_Vector2(point))
        self.update(append=True)

    def remove_point(self, index: int):
        if index >= len(self._points):
            raise IndexError(f"Point number {index} out of range.")
        else:
            self._points.pop(index)
            self.update()

    def set_point(self, index: int, point: Union[Tuple[float, float], sf.Vector2]):
        if index >= len(self._points):
            raise IndexError(f"Point number {index} out of range.")
        else:
            self._points[index] = to_Vector2(point)
            self.update()

    def update(self, append=False, color=False, offset=False):
        if append:
            self._vertexarray.append(sf.Vertex(position=self._points[-1], color=self._color))
        if color:
            for i in range(self.point_count):
                self._vertexarray[i].color = self._color
        if offset:
            for vertex in self._vertexarray:
                vertex.position += offset
        if not color and not append and not offset:
            self._vertexarray.resize(self.point_count)
            for i in range(self.point_count):
                self._vertexarray[i].position = self._points[i]
                self._vertexarray[i].color = self._color

    def draw(self, target, states):
        target.draw(self._vertexarray, states)


class EllipseShape(sf.ConvexShape):
    def __init__(self, radius: Union[Tuple[float, float], sf.Vector2]):
        super().__init__()
        self._radius = to_Vector2(radius)
        self.point_count = 30
        self.update()

    def update(self):
        for i in range(self.point_count):
            self.set_point(i, self._calc_point(i))

    def _calc_point(self, index: int):
        pi = 3.141592654
        angle = index * 2 * pi / self.point_count - pi / 2
        x = math.cos(angle) * self._radius.x
        y = math.sin(angle) * self._radius.y

        return sf.Vector2(self._radius.x + x, self._radius.y + y)

    def set_point_count(self, count: int):
        self.point_count = count
        self.update()

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, radius: Union[Tuple[int, int], sf.Vector2]):
        self._radius = to_Vector2(radius)
        self.update()



from typing import Dict, Union
import os

from sfml import sf

from ..tilemapping import tiledmap as tm
from .resource_path import find_resource
from .tileset_manager import TilesetManager


def load_resources(obj, path):
    for file in os.listdir(find_resource(path)):
        filename, ext = os.path.splitext(file)
        if os.path.isfile(os.path.join(path, file)):
            if ext in [".png", ".jpg", ".bmp"]:
                value = sf.Texture.from_file(find_resource(os.path.join(path, file)))
                setattr(obj, filename, value)
                res = obj
                while res._parent:
                    res = res._parent
                setattr(res._textures, filename, value)

            elif ext in [".ttf"]:
                value = sf.Font.from_file(find_resource(os.path.join(path, file)))
                value.get_texture(8).smooth = False
                value.get_texture(16).smooth = False
                value.get_texture(32).smooth = False
                setattr(obj, filename, value)
                res = obj
                while res._parent:
                    res = res._parent
                setattr(res._fonts, filename, value)

            elif ext in [".tmx"]:
                value = tm.TiledMap(find_resource(os.path.join(path, file)))
                setattr(obj, filename, value)
                res = obj
                while res._parent:
                    res = res._parent
                setattr(res._maps, filename, value)

            elif ext in [".tsx"]:
                TilesetManager.load_tsx(find_resource(os.path.join(path, file)))

        elif os.path.isdir(os.path.join(path, file)):
            directory = Dir(filename, parent=obj)
            load_resources(directory, os.path.join(path, file))
            setattr(obj, filename, directory)


class Dir:
    def __init__(self, name, parent=None):
        self._name = name
        self._parent = parent
        self._data: Dict[str, Union[Dir, sf.Texture, tm.TiledMap, sf.Font]] = {}

    def __getattr__(self, item) -> Union[sf.Font, sf.Texture, tm.TiledMap]:
        if item[0] == '_':
            return self.__dict__[item]
        if item == '..':
            return self._parent
        if item in self._data:
            return self._data[item]
        raise AttributeError(f"File or directory '{item}' not found in directory '{self._name}'")

    def __setattr__(self, key, value):
        if key not in vars(dict).keys():
            if key[0] == '_':
                self.__dict__[key] = value
            else:
                self._data[key] = value
        else:
            raise AttributeError(f"{key} is not a valid key")

    def __iter__(self):
        for key in self._data.keys():
            yield key

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    def all_files(self):
        for file in self.values():
            if isinstance(file, Dir):
                for f in file.all_files():
                    yield f
            else:
                yield file

    def print_tree(self, indent=0):
        items = [i for i in self.items()]
        dirs = [d for d in items if isinstance(d[1], Dir)]
        dirs.sort(key=lambda x: x[0])
        files = [f for f in items if not isinstance(f[1], Dir)]
        files.sort(key=lambda x: x[0])

        for name, val in dirs:
            print(("│" + " " * 3) * (indent - 1), end='')
            print("├" + "─", end='')
            print(f" {name} : {type(val).__name__}")
            val.print_tree(indent=indent + 1)

        for name, val in files:
            print(("│" + " " * 3) * (indent - 1), end='')
            if (name, val) == files[-1]:
                print("└" + "─", end='')
            else:
                print("├" + "─", end='')
            print(f" {name} : {type(val).__name__}")



from typing import List, Dict, Union, Optional
from xml.etree import ElementTree

from sfml import sf

from ..data.rect import Rect
from ..data.shapes import EllipseShape, LineShape
from .tiles import Tile


class TileLayer(sf.Drawable):
    def __init__(self, tiledmap, xml: ElementTree.Element):
        super().__init__()
        self.xml = xml
        self.map = tiledmap

        self.id: int = int(self.xml.get('id'))
        self.name: str = self.xml.get('name')
        self.width: int = int(self.xml.get('width'))
        self.height: int = int(self.xml.get('height'))
        self.visible: bool = not bool(self.xml.get('visible'))

        self.tiles: List[List[Optional[Tile]]] = []
        self.properties: Dict[str, Union[int, float, str, bool, sf.Color]] = {}
        self.animated_tiles: Dict[tuple, Tile] = {}

        self.render_texture: sf.RenderTexture = sf.RenderTexture(self.width*self.map.tile_width,
                                                                 self.height*self.map.tile_height)
        self.sprite: sf.Sprite = sf.Sprite(self.render_texture.texture)

    def parse(self):
        # parsing properties of the layer
        if self.xml.find('properties'):
            for prop in self.xml.find('properties'):
                name = prop.get('name')
                prop_type = prop.get('type')
                value = prop.get('value')
                if prop_type == "bool":
                    value = True if value == "true" else False
                elif prop_type == "int":
                    value = int(value)
                elif prop_type == "float":
                    value = float(value)
                elif prop_type == "color":
                    value = sf.Color(int(value[1:3], 16), int(value[3:5], 16), int(value[5:7], 16), int(value[7:9], 16))

                self.properties[name] = value

        self.render_texture.clear(sf.Color.TRANSPARENT)
        layer_data = [int(n) for n in self.xml.find('data').text.split(',')]
        # iterating over the tiles of the layer
        for j in range(self.height):
            self.tiles.append([])
            for i in range(self.width):
                gid = layer_data[i + j * self.width]
                if gid != 0:
                    tile = Tile(self, i, j, gid)
                    self.tiles[j].append(tile)
                    self.render_texture.draw(tile)
                else:
                    self.tiles[j].append(None)

        self.render_texture.display()
        self.sprite.texture = self.render_texture.texture

    @property
    def position(self):
        return self.sprite.position

    @property
    def y(self):
        return self.sprite.position.y

    def update_tile(self, i, j):
        tile = self.tiles[j][i]
        if tile is not None and tile.animated:
            tile.update()
            clear = sf.RectangleShape()
            clear.position = tile.position
            clear.size = tile.size
            clear.fill_color = sf.Color.TRANSPARENT
            self.render_texture.draw(clear, sf.RenderStates(sf.BLEND_NONE))
            self.render_texture.draw(tile)

    def draw(self, target, states):
        if self.visible:
            target.draw(self.sprite, states)
        else:
            print(f"Trying to draw invisible TileLayer {self.name}")


class ObjectGroup(sf.Drawable):
    def __init__(self, tiledmap, xml: ElementTree.Element):
        super().__init__()
        self.xml = xml
        self.map = tiledmap

        self.id = int(self.xml.get('id'))
        self.name = self.xml.get('name')

        color = self.xml.get('color')
        if color:
            if len(color) == 9:
                self.color = sf.Color(int(color[3:5], 16), int(color[5:7], 16), int(color[7:9], 16), int(color[1:3], 16))
            else:
                self.color = sf.Color(int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16), 255)
        else:
            self.color = sf.Color(100, 100, 100, 160)

        self.rectangles = []
        self.points = []
        self.ellipses = []
        self.polygons = []
        self.lines = []

        self._render_texture = sf.RenderTexture(tiledmap.width*tiledmap.tile_width, tiledmap.height*tiledmap.tile_height)
        self.sprite = sf.Sprite(self._render_texture.texture)

    def parse(self):
        self._render_texture.clear(sf.Color.TRANSPARENT)
        pts_vertexarray = sf.VertexArray(sf.PrimitiveType.POINTS)
        rect_vertexarray = sf.VertexArray(sf.PrimitiveType.QUADS)
        for object_elmnt in self.xml.findall('object'):
            x = float(object_elmnt.get('x'))
            y = float(object_elmnt.get('y'))
            width = float(object_elmnt.get('width')) if object_elmnt.get('width') else 0
            height = float(object_elmnt.get('height')) if object_elmnt.get('height') else 0
            if not list(object_elmnt):
                self.rectangles.append(Rect((x, y), (width, height)))
            else:
                child = list(object_elmnt)[0]
                if child.tag == "ellipse":
                    ellipse = EllipseShape((width/2, height/2))
                    ellipse.position = (x, y)
                    ellipse.fill_color = self.color
                    self._render_texture.draw(ellipse)
                elif child.tag == "point":
                    self.points.append(sf.Vector2(x, y))
                elif child.tag == "polyline":
                    points = []
                    for strpt in child.get('points').split():
                        points.append(sf.Vector2(float(strpt.split(',')[0]), float(strpt.split(',')[1])))
                    line = LineShape(*points)
                    line.position = (x, y)
                    line.color = self.color
                    self._render_texture.draw(line)
                elif child.tag == "polygon":
                    points = []
                    for strpt in child.get('points').split():
                        points.append(sf.Vector2(float(strpt.split(',')[0]), float(strpt.split(',')[1])))
                    polygone = sf.ConvexShape(len(points))
                    for i, p in enumerate(points):
                        polygone.set_point(i, p)
                    polygone.position = (x, y)
                    polygone.fill_color = self.color
                    self.polygons.append(polygone)
                    self._render_texture.draw(polygone)

        rect_vertexarray.resize(4*len(self.rectangles))
        for index in range(0, len(self.rectangles)*4, 4):
            rect_vertexarray[index + 0].position = self.rectangles[index//4].topleft
            rect_vertexarray[index + 1].position = self.rectangles[index//4].topright
            rect_vertexarray[index + 2].position = self.rectangles[index//4].bottomright
            rect_vertexarray[index + 3].position = self.rectangles[index//4].bottomleft
            for i in range(4):
                rect_vertexarray[index + i].color = self.color

        pts_vertexarray.resize(len(self.points))
        for index in range(len(self.points)):
            pts_vertexarray[index].position = self.points[index]
            pts_vertexarray[index].color = self.color

        self._render_texture.draw(rect_vertexarray)
        self._render_texture.draw(pts_vertexarray)
        self._render_texture.display()
        self.sprite.texture = self._render_texture.texture

    @property
    def position(self):
        return sf.Vector2(0, 0)

    def draw(self, target, states):
        target.draw(self.sprite)



import inspect
import os

from sfml import sf

from ..reslib import Res
from ..data.game_obj import GameObject


class Logger(GameObject):
    @classmethod
    def log(cls, *args):
        line = inspect.stack()[1][2]
        path = inspect.stack()[1][1].split(os.sep)
        file = path[-2]+'/'+path[-1]
        metadata = '[' + str(line) + '|' + file + ']'
        while len(metadata) < 25:
            metadata += ' '
        if cls.game.debug:
            print(metadata, end=" ")
            for arg in args:
                print(arg, end=" ")
            print()


class DebugText(sf.Text):
    def __init__(self, instance: object, attr_name: str, position: tuple, float_round=2):
        super().__init__()
        self.obj_instance = instance
        self.attr_name = attr_name
        self._float_round = float_round
        self.string = attr_name+" : "+str(getattr(instance, attr_name))
        self.font = Res.Fonts.arial
        self.character_size = 16
        self.color = sf.Color.WHITE
        self.position = position

    def update(self):
        value = getattr(self.obj_instance, self.attr_name)
        if isinstance(value, float):
            value = round(value, self._float_round)
        elif isinstance(value, sf.Vector2):
            value = (round(value.x, self._float_round), round(value.y,self._float_round))
        self.string = self.attr_name+" : "+str(value)




from .const import SCREEN_W, SCREEN_H
from .keys import Keyboard
from .rect import Rect
from .shapes import LineShape, EllipseShape

__all__ = ['Keyboard', 'Rect', 'LineShape', 'EllipseShape', 'SCREEN_W', 'SCREEN_H']



from setuptools import setup, find_packages

with open("readme.md", "r") as rdme:
    long_description = rdme.read()


setup(
    name="NasNas",
    version="0.1.3",
    author="Modar Nasser",
    author_email="modar1999@gmail.com",
    description="A simple game framework to get started quickly with python and sfml.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Madour/pyNasNas",
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    keywords=["sfml", "game dev", "game engine", "framework"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: zlib/libpng License",
        "Programming Language :: Python :: 3"
    ]
)



from .styles import ButtonStyle, MenuStyle, BoxBorder
from .buttons import Button
from .menus import Menu, VerticalMenu

__all__ = [
    'ButtonStyle', 'MenuStyle', 'BoxBorder',
    'Button',
    'Menu', 'VerticalMenu'
]



from typing import Dict
from xml.etree import ElementTree

from .resource_path import find_resource
from ..tilemapping.tilesets import Tileset


class TilesetManagerMeta(type):
    _data: Dict[str, Tileset] = {}

    @classmethod
    def __iter__(mcs):
        for x in mcs._data.values():
            yield x

    @classmethod
    def __contains__(mcs, item):
        if isinstance(item, str):
            if item in mcs._data:
                return True
            return False
        else:
            raise TypeError("Error, `in` comparison with TilesetManager should be a string")


class TilesetManager(metaclass=TilesetManagerMeta):
    @classmethod
    def load_tsx(cls, path: str):
        xml = ElementTree.parse(find_resource(path)).getroot()
        t = Tileset(xml, path)
        cls._data[t.name] = t

    @classmethod
    def get(cls, tilesetname: str):
        if tilesetname in cls._data:
            return cls._data[tilesetname]



from typing import Dict

from sfml import sf

from ..data.game_obj import GameObject
from . import layers


class Scene(GameObject, sf.Drawable):
    def __init__(self, width: int, height: int):
        super().__init__()
        self.render_texture: sf.RenderTexture = sf.RenderTexture(width, height)
        self.sprite: sf.Sprite = sf.Sprite(self.render_texture.texture)
        self.layers: Dict[int, layers.Layer] = {}
        self.masks: Dict[int, layers.Mask] = {}

    @property
    def width(self) -> int:
        return self.render_texture.width

    @property
    def height(self) -> int:
        return self.render_texture.height

    def add_layer(self, layer: layers.Layer, order: int):
        """ Adds a Layer to the scene at the given order
        Order 0 will be drawn first on the scene, then 1, then 2 etc
        """
        if isinstance(layer, layers.Layer):
            self.layers[order] = layer
        else:
            raise TypeError("Argument layer should be a Layer instance.")

    def remove_layer(self, layer: layers.Layer = None, order: int = None):
        if layer:
            if layer in self.layers.values():
                self.layers = {key: val for key, val in self.layers.items() if val != layer}
        elif order:
            if order in self.layers.keys():
                self.layers.pop(order)

    def get_layer(self, name: str = None, order: int = None) -> layers.Layer:
        for layer_order, layer in self.layers.items():
            if name is not None and order is not None:
                if self.layers[order].name != name:
                    raise IndexError(f"Requested order {order} and name {name} does not match")
            if name is not None:
                if layer.name == name:
                    return layer
            if order is not None:
                if layer_order == order:
                    return layer

    def add_mask(self, mask: layers.Mask, order: int):
        if isinstance(mask, layers.Mask):
            self.masks[order] = mask
        else:
            raise TypeError("Argument mask should be a Mask instance.")

    def remove_mask(self, mask: layers.Mask):
        if mask in self.masks.values():
            self.masks = {key: val for key, val in self.masks.items() if val != mask}

    def render(self):
        self.render_texture.clear(sf.Color.TRANSPARENT)

        max_layers_order = max(self.layers.keys()) if self.layers.keys() else 0
        max_masks_order = max(self.masks.keys()) if self.masks.keys() else 0

        for i in range(max(max_layers_order, max_masks_order)+1):
            if i in self.layers:
                self.layers[i].update()
                self.render_texture.draw(self.layers[i])
            if i in self.masks:
                self.masks[i].update()
                self.render_texture.draw(self.masks[i])

        self.render_texture.display()
        self.sprite.texture = self.render_texture.texture

    def draw(self, target, states):
        target.draw(self.sprite)
